/*
 macos/macsound.c
  sound for MacFCEU

  Created: 2008.07.02
  Revised: 2012.02.16

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Tuesday, September 16, 2008 10:52 PM EDST
  because playback end starts at buffered end, double buffering can be removed
Tuesday, January 19, 2009 11:43 PM
  sound rate NTSC mod (ConsoleDownRate) moved to sound.c.
Saturday, October 24, 2009 9:08 PM
  officially getting rid of the non-68k callback stuff.
*/

#include <stdio.h>

#include <Sound.h>

#include "shell.h"

#include "../FCEU.h"
#include "../utility/ftrig.h"
#include "../common/dface.h"
#include "../WAVE.h"  /* for WAVE_DefaultRate */

#include "shellmenus.h"

#ifdef env68k
 #pragma segment Sound
#endif

#ifndef NO_DIRECT_COMPONENT_SOUND
 #define NO_DIRECT_COMPONENT_SOUND	1
#endif

/* ---- globals */

/* surpassing buffer size in sound.c */
#define LOCAL_BUFFER_LENGTH		((12<<10)+16)
#define LOCAL_BUFFER_MAXIMUM		(LOCAL_BUFFER_LENGTH-32)

static int16		playbuffer[LOCAL_BUFFER_LENGTH*2];
static int		bufferselect = 0; /* double-buffered this time; ^=LOCAL_BUFFER_LENGTH */

static SndCallBackUPP	userroutine;

/* ---- macros */

#if TARGET_CPU_68K
 #define __FIXEDEXT80(x)		((x)/65536.0)
#else
 #define __FIXEDEXT80(x)		{0x400E, x, 0}
#endif

/* ---- code */

SndChannel	mychannel =
{
	NULL, NULL, NULL, /* next channel, first mod, callback proc */
	0, 0, /* user info,  wait */
	{0}, /* command in progress */
	0, stdQLength, 0, 0, /* flags, qLength, qHead, qTail */
};
SndChannel	*mychannelptr = &mychannel;

#if !NO_DIRECT_COMPONENT_SOUND
 /* hack! */
 #define GetSCComponent(c)		(0+*(ComponentInstance*)(c)->firstMod)
 /* warning: could not +0 (Opaque structure) */
 #define GetSCSource(c)			(*(SoundSource*)(16+(UInt8*)(c)->firstMod))

static pascal Boolean myrebuffer(SoundParamBlock **block)
{ /* swap buffer block and play if filled */
	(**block).desc.sampleCount = 0; /* let another fill */

	return false;
}
/* sound will only break up if next concurrent block isnt already filled */
static pascal Boolean myreplay(SoundParamBlock **block)
{ /* play next filled buffer, if any */
	SoundParamBlock		*param = *block;

	--mychannel.userInfo;
	return param->desc.sampleCount!=0;
}

static SoundParamBlock	myblock =
{
	sizeof(SoundParamBlock), /* structure size */
	0, BIG_ENDIAN ? k16BitBigEndianFormat:k16BitLittleEndianFormat, /* flags, format */
	1, 16, /* channel count, sample size */
	rate22khz, /* sample rate */
	0, NULL, /* sample count, buffer */
	0, 0x10000L, /* reserved, rate multiplier */
	0x100, 0x100, /* short-fixed volume (left, right) */
	kBestQuality, /* interpolation is recommended */
	0, /* filter */
	myrebuffer, /* rebuffer routine */
	myreplay, /* completion routine */
	0, /* playing */
	0
};
#endif

#if BIG_ENDIAN
ExtSoundHeader		myheader =
{
	NULL, /* sample data */
	1, rate22khz, 0, 0, /* channel count, sample rate, loop start, end */
	extSH, 60, /* encoding, base note */
	0, __FIXEDEXT80(rate22khz), /* frame count, floating sample rate */
	NULL, NULL, NULL, /* marker sync track, instrument tracks, AESRecording */
	16, /* sample size */
};
#else
CmpSoundHeader		myheader =
{
	NULL, /* sample data */
	1, rate22khz, 0, 0, /* channel count, sample rate, loop start, end */
	cmpSH, 60, /* encoding, base note */
	0, __FIXEDEXT80(rate22khz), /* frame count, floating sample rate */
	NULL, /* marker sync track */
	k16BitLittleEndianFormat, /* format */
	0, NULL, NULL, 0, /* reserved, state block ptr, leftover samples, codec ID */
	0, 0, 16, /* packet size, synth ID, sample size */
};
#endif

SndCommand	bfcommand =
{
	bufferCmd, 0, (long int)&myheader
};
SndCommand	cbcommand =
{
	callBackCmd, 0, 0
};

int KillSound()
{
	if (!mychannel.firstMod/* || !soundavailable*/)
		return 0;
	myheader.samplePtr = (Ptr)
"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";
	myheader.numFrames = 32;

	mychannel.userInfo = 1;
	SndDoImmediate(&mychannel,&bfcommand);
	/*do{TickCount();} while(mychannel.userInfo==0);*/
	mychannel.userInfo = 0;

	SndDisposeChannel(&mychannel,true);
	mychannel.firstMod = NULL;

#if CALL_NOT_IN_CARBON
	UnlockMemory(playbuffer,sizeof(playbuffer));
#endif
	DisposeSndCallBackUPP(userroutine); userroutine = NULL;

	return 1;
}

static pascal void mychannelcb(SndChannelPtr chan,
	SndCommand *cmd)
{
#pragma unused(cmd)
//extern int32 frametime;
	if (--chan->userInfo<0)
		mychannel.userInfo = 0;
}

/* null git now taken as normal */
/* (soundon/soundvolume determine if channel should be created) */
int InitSound(FCEUGI *git)
{
	UnsignedFixed	samplerate, ratemultiplier = 0x10000UL;
	int		stereo;
	OSErr		err;

	if (mychannel.firstMod)
		KillSound();

	if (!soundon || !soundvolume) /*git==NULL)*/
		{FCEUI_Sound(0); return 0;}

	/* fix sample rate variables */
	if (!git)
		samplerate = soundrate;
	else if ((samplerate=git->soundrate)==0)
		git->soundrate = samplerate = soundrate;
	else if (soundrate==0)
		soundrate = samplerate;

	if (samplerate<1000 || samplerate>96000)
	{ /* fix range */
		samplerate = soundrate = samplerate<1000 ? 1000:96000;
		updatemenus();
		FCEUI_SetSoundQuality(0);
	}

	stereo = git && git->soundchan==2 ? 1:0;

	FCEUI_Sound(WAVE_DefaultRate=samplerate);
	FCEUI_SetSoundActiveChannels(sndactivechannels);

	if (!soundavailable)
		return 1;
#if CALL_NOT_IN_CARBON
	LockMemory(playbuffer,sizeof(playbuffer));
#endif
	err =-1;
	if ((userroutine=NewSndCallBackUPP(mychannelcb))==NULL
			|| (err=SndNewChannel(&mychannelptr,sampledSynth,stereo ?
			initStereo:initMono,userroutine))!=noErr)
	{
		FCEU_printf("## InitSound: failed to create sound channel (#%d)\n",
			err);
		return -1;
	}

	{ /* support for sample rates >= 65536 */
		int		shift;

		if ((shift=cntlz32(samplerate))>15)
			samplerate<<= 16;
		else if ((ratemultiplier<<=16+-shift)==0)
			{ratemultiplier =-1UL; /*DebugStr("\pDemon, be Gone!");*/}
		else
			samplerate<<= shift;
	}
#if !NO_DIRECT_COMPONENT_SOUND
	myblock.desc.numChannels = 1+stereo;
	myblock.desc.sampleRate = samplerate;
	myblock.rateMultiplier = ratemultiplier;
	myblock.refCon = 0;
#else
	myheader.numChannels = 1+stereo;
	myheader.sampleRate = samplerate;
	if (ratemultiplier!=0x10000UL)
	{
		SndCommand	command;

		command.cmd = rateMultiplierCmd;
		command.param1 = 0;
		command.param2 = ratemultiplier;
		err = SndDoImmediate(mychannelptr,&command);
	}
#endif
	// throttle always o now; 2012.01.03 04:46 ET
	RefreshThrottleFPS();

	return 1;
}

void writesound(int16 *data,
	int count)
{
	if (!mychannel.firstMod || count<=0 || !data)
		return;
	if (mychannel.userInfo) /* this used to be at the end */
	{ /* wait till prev. is finished */
		/* overflow protected: 1/30-4.25 sec. */
		int	ticks = 2+TickCount();

		do /* this has still been known to overflow until reset */
			if (!mychannel.userInfo)
				break;
		while((int8)(TickCount()-ticks)<0);
	}
	/* copy low 4 bits, +16 for memcpy16 */
	myheader.samplePtr = (Ptr)((15&(int)data)+16
		+(~15&(bufferselect+(long int)playbuffer)));
	bufferselect^= LOCAL_BUFFER_LENGTH; /* alternate */

	if (count>LOCAL_BUFFER_MAXIMUM)
		count = LOCAL_BUFFER_MAXIMUM;
	memcpy16(myheader.samplePtr,data,count*2+16);
#if !NO_DIRECT_COMPONENT_SOUND
	if (NoWaiting) /* clear for immediate playback */
		myblock.desc.sampleCount = 0;

	myblock.desc.buffer = (UInt8*)myheader.samplePtr;
	myblock.desc.sampleCount = count;

	if (!myblock.desc.sampleCount) //((wait=..!=0)==0)
		SoundComponentPlaySourceBuffer(GetSCComponent(&mychannel),
			GetSCSource(&mychannel),&myblock,0);
#else
	myheader.numFrames = count;
	if (NoWaiting&1)
	{
		SndDoImmediate(&mychannel,&bfcommand);
		mychannel.userInfo = 0;
	} else
	{
		mychannel.userInfo+= 1; //!NoWaiting;

		SndDoCommand(&mychannel,&bfcommand,0); //NoWaiting);
		SndDoCommand(&mychannel,&cbcommand,0);
	}
#endif
}

void stopsound()
{
	if (!mychannel.firstMod)
		return;
#if !NO_DIRECT_COMPONENT_SOUND
	SoundComponentStopSource(GetSCComponent(&mychannel),1,&GetSCSource(&mychannel));
	myblock.desc.sampleCount = 0;
#else
	{
		SndCommand	command;

		command.cmd = quietCmd;
		command.param1 = 0;
		command.param2 = 0;
		SndDoImmediate(&mychannel,&command);
		command.cmd = flushCmd;
		SndDoImmediate(&mychannel,&command);
	}
#endif
	mychannel.userInfo = 0;
}

boolean adequatesoundmanager()
{
	NumVersion	version = SndSoundManagerVersion();

	return version.majorRev>=3;
}

int adjust_soundvolume(int displacement)
{
	int		volume = soundvolume/10; /* shouldnt directly set the global */

	if (!soundon)
		return -1;
	if ((volume+=displacement)<0)
		volume = 0;
	else if (volume>15)
		volume = 16;

	return domenu(_subvolume_off+volume); /* should call InitSound during 0% volume game */
}
