/*
 macos/messagelog.c
  and not to stderr... (no longer stderr)

  Created: 2008.07.12
  Revised: 2011.03.24

Sunday, July 20, 2008 7:48 PM EDST
  rewrote list mechanisms.
  free and malloc should be realloc(..,0) and realloc(NULL,..) to minimize code space.
Monday, April 5, 2010 1:54 PM
  removed unused setlogwindowscrolls.
  logwindow_scroll now caps to technical Classic limits (as it should've for a long time),
 also multiplies by magnitude only if applied (INT_NO_VALUE must be preserved if not).
*/

#include <stdio.h>
#include <stdlib.h>

#include <MacWindows.h>
#include <Events.h>
#include <Controls.h>
#include <ControlDefinitions.h>

#include "shell.h"
#include "macvideo.h"
#include "messagelog.h"

/* ---- constants */

#define PASCAL_LOG_FONT_NAME		"\pMonaco"

#define LOGTEXT_INDENT_H		6
#define LOGTEXT_INDENT_V		0

#define MAXIMUM_LINESIZE		512 /* should be <=16000 */
#define MAXIMUM_LINECOUNT		150 /* 130 can dm all RAM; was 64, 100, 128, 130 */
#define MAXIMUM_LOGSIZE			(512*1024) /* was 64KB */

#define CHR_LF				'\012'
#define CHR_VT				'\013'
#define CHR_FF				'\014'
#define CHR_CR				'\015'

#define INIT_LOGWINDOW_WIDTH		480
#define INIT_LOGWINDOW_HEIGHT		(8*12+18)

#define LOGWINDOW_MINIMUM_HEIGHT	(INIT_LOGWINDOW_HEIGHT-24)
#define LOGWINDOW_MINIMUM_WIDTH		196
#define LOGWINDOW_MAXIMUM_HEIGHT	(MAXIMUM_LINECOUNT/2*12+16)
#define LOGWINDOW_MAXIMUM_WIDTH		(MAXIMUM_LINESIZE/2*6)

/* ---- globals */

WindowRef		logwindow = 0;

/* note that a message log window can exist while the logfile doesnt have to */
static FILE		*logfile = 0;
/* +1 for line pointer list for final length, +1 for termination */
static char		*logtext;
static long		logtextlength;
static const char	*loglinelist[MAXIMUM_LINECOUNT+3];
static const char	**logfirstline = 1+loglinelist;
static const char	**logfirstindex = 1+loglinelist; /* elderscroll */
static int		logcount = 0, loglineheight = 12;/*, viewheight = 64;*/

static boolean		fullupdate = false;
unsigned char		logtofile = 1; /* *512KB clipped */

/* ---- macros */

#define ShowHideControl(c,x)		do{if(x) ShowControl(c); else HideControl(c);}while(0);

#if !TARGET_API_MAC_CARBON
 #define IsControlVisible(h)		(0+(**(h)).contrlVis)
#endif

/* ---- code */

/* appending text to end requires that we scan in reverse (turrible) */
/* an extra entry beyond count is required */
/* returns the earliest line pointer (for map) */
static const char **rcalclines(const char **list,
	int *count,
	const char *text,
	long length)
{
	const char	*s;
	long		t = 1, x;
	uint8		p = 0xFF;

	if (!count)
		return 0;
	x = *count;
	*count = 0;

	if (!list || (s=text)==NULL || length<=0)
		return 0;

	s+= length;
	list+= x;
	*list = s;

	do
	{
		uint8		c = *--s;

		if ((c+=-CHR_LF)<=CHR_CR-CHR_LF) /* any EOL */
		{
			if (p==0 && c==(uint8)(CHR_CR-CHR_LF)) /* DOS line ending (skip) */
				{p = 0xFF; continue;}
			++t;
			*--list = 1+s;
			if (--x==0)
				{++s; break;}
		}
		p = c;
	} while(--length!=0);

	*--list = s;
	/* all eols need to be accounted for (count a final empty line) for erasure */
	*count = t;

	return list;
}

static void getlogwindowscrolls(int *h,
	int *v)
{
	ControlRef	control;

	if (v && (control=*(ControlRef*)(240+(Ptr)logwindow))!=NULL)
		*v = GetControlValue(control);
	if (h && (control=*(ControlRef*)(248+(Ptr)logwindow))!=NULL)
		*h = GetControlValue(control);
}

static void drawlogwindowtext(int c)
{
	const char	**a;
	int		i = 0, h, x = 0, y;

	getlogwindowscrolls(&x,&i);
	x*= -6;
	x+= LOGTEXT_INDENT_H;
	y = LOGTEXT_INDENT_V+loglineheight;

	/* because of font ascent, include previous line (if it exists) */
	if (i>0)
		--i, y-= loglineheight, ++c;

	a = i+logfirstindex;

	if (!logtext || (unsigned)i>=MAXIMUM_LINECOUNT)
		return;
	if ((unsigned)(h=logcount-i)>c)
		h = c;
	do
	{
		long		z = a[1]-*a;
		const char	*s;

		if ((s=a[0])==NULL)
			break; /* termination check */
		MoveTo(x,y);
		if (z>0 && s[z-1]=='\012') /* dont draw LF; 2009.06.27 13:47 EST */
			--z;
		DrawText(s,0,z>MAXIMUM_LINESIZE*2 ? 1+MAXIMUM_LINESIZE*2:z);
	} while(++a,y+=loglineheight,--h!=0);
}

static void logwindow_scrollbox(short sh,
	short sv)
{
	Rect		newfocus;
	RgnHandle	region = NewRgn();
	GrafPtr		oldport;

	GetPort(&oldport);
	setportwindow(logwindow);

	getwindowbounds(logwindow,&newfocus);
	newfocus.bottom+= -16;
	newfocus.right+= -16;

	ScrollRect(&newfocus,sh,sv,region);

	if (region)
	{ /* update what ScrollRect has given us (the remainder that needs rendering) */
		SetClip(region);
		drawlogwindowtext((newfocus.bottom-newfocus.top)/loglineheight);

		DisposeRgn(region);

		/* reset clip (anything large enough will do) */
		*(SInt32*)&newfocus.top = 0;
		*(SInt32*)&newfocus.bottom = 0x40004000;
		ClipRect(&newfocus);
	}

	MacSetPort(oldport);
}

void logwindow_scroll(int hcount,
	int vcount)
{
	ControlRef	control;
	int		hsep = INT_NO_VALUE, vsep = INT_NO_VALUE;
	int		vshift = vcount&(cmdKey<<16) ? 11:0;

	if (hcount>0xFFFF)
	{
		if ((hcount&=0xFF)==0)
		{ /* command page -> vertical to horizontal scroll */
			if ((hcount=vcount&0xFF)>kPageDownCharCode || !vshift)
				{hcount = 0; goto skip;}
			vcount = 0;
		}

		switch(hcount)
		{
		case kPageUpCharCode: hcount = -8; break;
		case kPageDownCharCode: hcount = 8; break;
		case kUpArrowCharCode: hcount = -1; break;
		case kDownArrowCharCode: hcount = 1; break;
		default: hcount = 0;
		}
	}

	if (vcount>0xFFFF)
		switch(vcount&0xFF)
		{
		case kPageUpCharCode: vcount = -8<<vshift; break;
		case kPageDownCharCode: vcount = 8<<vshift; break;
		case kUpArrowCharCode: vcount = -1<<vshift; break;
		case kDownArrowCharCode: vcount = 1<<vshift; break;
		default: vcount = 0;
		}
skip:
	if ((control=*(ControlRef*)(240+(Ptr)logwindow))!=NULL)
	{
		vsep = GetControlValue(control);
		if ((vcount+=vsep)<-32768)
			vcount = -32768;
		else if (vcount>32767)
			vcount = 32767;
		SetControlValue(control,vcount);
		vsep-= GetControlValue(control);
		vsep*= loglineheight;
	}
	if ((control=*(ControlRef*)(248+(Ptr)logwindow))!=NULL)
	{
		hsep = GetControlValue(control);
		if ((hcount+=hsep)<-32768)
			hcount = -32768;
		else if (hcount>32767)
			hcount = 32767;
		SetControlValue(control,hcount);
		hsep-= GetControlValue(control);
		hsep*= 6;
	}

	logwindow_scrollbox(hsep,vsep);
	fullupdate = true;
}

static int updatelogwindowscrolls(boolean added)
{
	Rect		bounds;
	ControlRef	control;
	int		value, vertc = 64;

	if ((control=*(ControlRef*)(240+(Ptr)logwindow))!=NULL)
	{ /* vertical */
		GetControlBounds(control,&bounds);
		if ((vertc=logcount-(bounds.bottom-bounds.top-15)/loglineheight-1)<=0)
			SetControlMaximum(control,0);
		else
		{ /* set maximum, update scroll position if at maximum */
			value = GetControlMaximum(control);

			SetControlMaximum(control,vertc);

			if (added && GetControlValue(control)==value) /*value>0 && */
			{
				SetControlValue(control,vertc);
				value-= vertc;
				logwindow_scrollbox(0,value*loglineheight);
			}
		}
	}

	if ((control=*(ControlRef*)(248+(Ptr)logwindow))!=NULL)
	{ /* horizontal */
		GetControlBounds(control,&bounds);
		value = MAXIMUM_LINESIZE-(bounds.right+bounds.left/6)+6;
		SetControlMaximum(control,value<0 ? 0:value);
	}

	return vertc;
}

static pascal void mycontrolactionproc(ControlRef control,
	ControlPartCode part)
{
	unsigned long	ticks = TickCount();
	unsigned long	reference = GetControlReference(control);
	SInt16		value;

	if (ticks<reference)
		return;
	value = GetControlValue(control);
	switch(part)
	{
	case kControlUpButtonPart:
		if (value>0)
			SetControlValue(control,value-1);
		break;
	case kControlDownButtonPart:
		if (value<GetControlMaximum(control))
			SetControlValue(control,value+1);
		break;
	case kControlPageUpPart:
		SetControlValue(control,value-8);
		break;
	case kControlPageDownPart:
		SetControlValue(control,value+8);
		break;
	case kControlIndicatorPart:
		break;
	}

	if ((value-=GetControlValue(control))!=0) /* sub-ing new; negative = forward */
	{ /* scroll window and update */
		GetControlTitle(control,(uint8*)temporary);
		if (temporary[1]=='V')
			logwindow_scrollbox(0,value*loglineheight);
		else
			logwindow_scrollbox(6*value,0);
	}

	if (reference==0)
		SetControlReference(control,ticks+30); /* 30/60 sec. delay */
}

static int calclineheight(void)
{
	FontInfo	fontinfo;
	SInt16		fontid;
	int		lh;

	GetFNum(PASCAL_LOG_FONT_NAME,&fontid);
	TextFont(fontid);
	TextSize(9);
	TextMode(srcCopy);
	GetFontInfo(&fontinfo);
	if ((lh=fontinfo.ascent+fontinfo.descent+1)==0)
		lh = 12;

	return lh;
}

static ControlActionUPP		mycontrolactionupp = 0;

static WindowRef createlogwindow(WindowRef behind)
{
	Str255		title;
	Rect		bounds;
	WindowRef	window;
	GrafPtr		oldport;
	Ptr		storage;

	GetPort(&oldport);

	/* bottom-center of screen */
	bounds = (**gamegdevice).gdRect;
	bounds.top = (bounds.bottom-=9)-INIT_LOGWINDOW_HEIGHT;
	bounds.left = bounds.left+bounds.right-INIT_LOGWINDOW_WIDTH>>1;
	bounds.right = bounds.left+INIT_LOGWINDOW_WIDTH;
	if ((storage=NewPtr(256))==NULL)
		return 0;
	*(uint32*)(252+storage) = 0;
	getpstr(title,strlistid_names,namesstr_logfilename);
	if ((window=NewCWindow(storage,&bounds,title,
			false,documentProc,behind,true,'mlog'))==NULL)
	{
		DisposePtr(storage);
		return 0;
	}

	setportwindow(window);

	loglineheight = calclineheight();

	bounds.top = -1;
	bounds.left = INIT_LOGWINDOW_WIDTH-15;
	bounds.bottom = INIT_LOGWINDOW_HEIGHT-14;
	bounds.right = INIT_LOGWINDOW_WIDTH+1;
	*(ControlRef*)(240+storage) = NewControl(window,&bounds,"\pV",true,
		0,0,0,scrollBarProc,0);

	bounds.top = INIT_LOGWINDOW_HEIGHT-15;
	bounds.left = -1;
	bounds.bottom = INIT_LOGWINDOW_HEIGHT+1;
	bounds.right = INIT_LOGWINDOW_WIDTH-14;
	*(ControlRef*)(248+storage) = NewControl(window,&bounds,"\pH",true,
		0,0,MAXIMUM_LINESIZE-64,scrollBarProc,0);

	MacSetPort(oldport);

	return window;
}

static int appendlogwindow(const void *str,
	int length)
{
	void		*t;

	if (logcount>=MAXIMUM_LINECOUNT)
	{
		int		diff;

		fullupdate = true;
		if (!logfirstindex)
			return -1;
		if (!*logfirstindex)
			++logfirstindex;
		if (*logfirstindex)
		{
			if ((diff=logtext-*logfirstindex)<0)
				diff = -diff;
			if ((logtextlength-=diff)<0)
				logtextlength = 0;
			else
				memmove(logtext,*logfirstindex,logtextlength);
		}
	}

	if ((t=realloc(logtext,logtextlength+length))==NULL)
		return -(length!=0);
	logtext = t;

	memcpy(logtextlength+logtext,str,length);

	logtextlength+= length;
	logcount = MAXIMUM_LINECOUNT;
	logfirstindex = rcalclines(logfirstline,&logcount,logtext,logtextlength);

	return 1;
}

void logwindow_clear()
{
	if (logtext)
		logtext = realloc(logtext,logtextlength=0);
	logcount = 0;
	loglinelist[MAXIMUM_LINECOUNT] = *logfirstline = 0;

	updatelogwindowscrolls(false);
	logwindow_update(true);
}

int logmessage(const char *str,
	long length)
{
	long		position;
	int		result = 0;

	if (length<0)
		length = strlen(str);
	if (!logwindow && (logwindow=createlogwindow((WindowRef)-1))==NULL)
		return -1;

	if (!logfile && logtofile)
	{
		getcstr(temporary,strlistid_names,namesstr_logfilename);
		if (temporary[0]) /* option: no log file by removing string (or strlist) */
			logfile = fopen(temporary,"w"); /* may fail (especially on a locked vol.) */
	}

	if (logfile && (uint32)(logtofile-1)>1
			&& (position=ftell(logfile))+length>=logtofile*MAXIMUM_LOGSIZE)
		fseek(logfile,0,SEEK_SET); /* overwriting self! */

	appendlogwindow(str,length);
	updatelogwindowscrolls(true); //viewheight = 
	logwindow_update(fullupdate);
	fullupdate = false;

	if (logfile && logtofile)
		result = fwrite(str,1,length,logfile);

	return result;
}

int logwindow_show(int w)
{
	if (!logwindow && (logwindow=createlogwindow((WindowRef)-1))==NULL)
		return -1;
	HiliteWindow(logwindow,true);

	if (w<0)
		w = !getvisiblewindowport(logwindow);
	else
		w = w!=0;

	ShowHide(logwindow,w);
	HiliteWindow(logwindow,true);

	/* otherwise, youd have to do window mode; 2009.09.27 09:11 EDST */
	if (w&fullscreen)
		BringToFront(logwindow);

	return w;
}

void logwindow_update(boolean erase)
{
	Rect		bounds;
	GrafPtr		oldport;

	if (!getvisiblewindowport(logwindow))
		return;
	getwindowbounds(logwindow,&bounds);

	GetPort(&oldport);
	setportwindow(logwindow);

	ClipRect(&bounds);
	DrawControls(logwindow);

	bounds.bottom-= 15;
	bounds.right-= 15;
	ClipRect(&bounds);

	if (erase)
		EraseRect(&bounds);

	--bounds.bottom, --bounds.right; /* not prep'd for full inset (drawing errors) */
	ClipRect(&bounds);

	drawlogwindowtext((bounds.bottom-bounds.top)/loglineheight);

	bounds.bottom+= 16, bounds.right+= 16;
	ClipRect(&bounds);

	if (erase==true)
		drawgrowicon(logwindow);

	MacSetPort(oldport);
}

static void logwindow_updatescrollpositions(void)
{
	Rect		bounds;
	ControlRef	control;
	int		right, bottom;

	if (!getvisiblewindowport(logwindow))
		return;
	getwindowbounds(logwindow,&bounds);
	right = bounds.right, bottom = bounds.bottom;

	if ((control=*(ControlRef*)(240+(Ptr)logwindow))!=NULL)
	{ /* vertical */
		bounds.top = -1;
		bounds.left = right-15;
		bounds.bottom = bottom-14;
		bounds.right = right+1;
		SetControlBounds(control,&bounds);
	}
	if ((control=*(ControlRef*)(248+(Ptr)logwindow))!=NULL)
	{ /* horizontal */
		bounds.top = bottom-15;
		bounds.left = -1;
		bounds.bottom = bottom+1;
		bounds.right = right-14;
		SetControlBounds(control,&bounds);
	}
}

static int logwindow_grow(const EventRecord *event)
{
	Rect		bounding;
	long		newdimensions;

	if (!getvisiblewindowport(logwindow))
		return -1;

	/* GrowWindow subtracts one when min=max, can adjust wholey via command key */
	bounding.top = LOGWINDOW_MINIMUM_HEIGHT;
	bounding.left = LOGWINDOW_MINIMUM_WIDTH;
	bounding.bottom = LOGWINDOW_MAXIMUM_HEIGHT;
	bounding.right = LOGWINDOW_MAXIMUM_WIDTH;
	if ((newdimensions=GrowWindow(logwindow,event->where,&bounding))!=0)
	{
		*(uint32*)&bounding = 0;
		bounding.right = (SInt16)newdimensions;

		/* make height divisible by the line height */
		newdimensions+= 6+(loglineheight/2)<<16;
		newdimensions/= loglineheight<<16;
		newdimensions*= loglineheight;
		bounding.bottom = (newdimensions-=6);

		SizeWindow(logwindow,bounding.right,bounding.bottom,false);
		logwindow_updatescrollpositions();
		updatelogwindowscrolls(false); /* this was missing for so long!; 2011.03.24 */
		/* because of bugs in SizeWindow, invalidation should be done manually */
		logwindow_update(true);

		return 1;
	}

	return 0;
}

void logwindow_click(EventRecord *event,
	short what)
{
	Point		hitpoint = event->where;
	ControlRef	control;

	ControlPartCode		part;

	if (!getvisiblewindowport(logwindow))
		return;
	localwindowpoint(logwindow,&hitpoint);

	if (what==inGrow)
		{logwindow_grow(event); return;}

	if ((control=*(ControlRef*)(240+(Ptr)logwindow))!=NULL && !IsControlVisible(control))
		logwindow_activate(1); /* manual update; 2010.05.27 */

	if (mycontrolactionupp==NULL
			&& (mycontrolactionupp=NewControlActionUPP(mycontrolactionproc))==NULL)
		return;
	if ((part=FindControl(hitpoint,logwindow,&control))==kControlNoPart || !control)
		return;

	SetControlReference(control,0); /* initialize timing for control */

	switch(part)
	{
	case kControlUpButtonPart:
	case kControlDownButtonPart:
	case kControlPageUpPart:
	case kControlPageDownPart:
		TrackControl(control,hitpoint,mycontrolactionupp);
		break;
	case kControlIndicatorPart:
		TrackControl(control,hitpoint,NULL);
		logwindow_update(2);
	}
}

void logwindow_activate(boolean activate)
{
	ControlRef	control;

	if (!getvisiblewindowport(logwindow))
		return;

	if ((control=*(ControlRef*)(240+(Ptr)logwindow))!=NULL)
		ShowHideControl(control,activate);
	if ((control=*(ControlRef*)(248+(Ptr)logwindow))!=NULL)
		ShowHideControl(control,activate);

	drawgrowicon(logwindow);

	HiliteWindow(logwindow,activate);
}
