/*
 utility/asm6502.c
  no globals stored here

 * MacFCEU 0.6a4
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>

#include "../types.h"
#include "asm6502.h"
#include "disasm6502.h"
#include "interpreter.h"

/* returns ID offset, ~ of end offset or ~0 (error) */
int32 X6502_AsmGetID(const char *s,
	const char *e,
	const _asm6502frame *f)
{
	uint8		n[256]; //max. size
	const uint8	*l;
	int		sz, z;
	uint8		c, h;

	if (f==0 || s==0 || e<=s)
		return -1;
	if ((l=f->labels)==0)
		return 0;
	h = 0;
	do
		h+= n[++sz-1] = toupper(*s++);
	while(s<e);
	h|= 0x80;

	for(; (c=*l++)!=0; l+=z)
	{
		if ((z=*l++)==0)
			z = 256;
		if (c!=h || sz!=z)
			continue;
		/* hash and type match */
		if (memcmp(l,n)==0)
			return ~(l-f->labels);
	}

	return l-f->labels;
}

/* returns ID offset, ~ of end already existing offset or ~0 (error) */
int32 X6502_AsmAddID(const char *s,
	const char *e,
	_asm6502frame *f)
{
	uint8		*d, h;
	int32		x;

	if ((x=X6502_AsmGetID(s,e))<=0)
		return x; /* already exists, or empty/error */
	if ((f->labels=realloc(f->labels,e-s+3))==0)
		return -2;
	d = 1+x+f->labels;
	*d++ = e-s;
	h = 0;
	do
		h+= *d++ = toupper(*s++);
	while(s<e);
	*d = 0;
	h|= 0x80;
	f->labels[x] = h;

	return 1;
}

static int getNameIndex(const char *s,
	const char *e)
{
	char		n[4];
	const char	*t;

	if (e-s!=3)
		goto nope; /* macros not handled */
	i = 0;
	do
		n[++i-1] = toupper(*s++);
	while(s<e);

	t = x6502_opnametable;
	i = 0;
	do
		if (*t==*n && t[1]==n[1] && t[2]==n[2])
			return i;
	while(t+=3,++i<OPN_COUNT);
nope:
	return -4;
}

int X6502_Assemble(const char *s,
	int (*writef)(uint16 addr, uint8 byte),
	_asm6502frame *f)
{
	_asm6502frame	af;
	const char	*e;
	int		r;
	uint8		c;

	if (s==0)
		return -1;
	r = 0;
	if ((e=strlen(s)+s)==s)
		goto end; /* empty */
	if (f==0)
	{
		f = &af;
		f->org = f->addr = 0; /* clear */
		f->flag = 0;
	}

	while((c=*s)!=0)
	{
		const char	*n;
		int		x;

		if (c=='\n' || c=='\r') /* new line */
		{
			f->flag&=~_IntoLine, ++s;
			continue;
		}
		if ((n=parse_skipwsp(s))==s && !(f->flag&_IntoLine))
		{ /* identifier: equate/label */
			n = parse_wordbreak(s,e);
			if ((r=X6502_AsmAddID(s,n,f))!=0)
				break;
		}
		f->flag|= _IntoLine;
		if ((n=parse_wordbreak(s,e))==s)
			{r = -3; break;} /* couldn't break - syntax error */
		/* macro or instruction */
		if ((x=getNameIndex(s,n))<0)
			{r = x; break;} /* can't identify */
		s = parse_skipwsp(n);

		/* parameters based on this, whatever it is... */
		//x6502_opn_table
	}

	if (f==&af)
	{
	finish:
		if (f->labels)
			{free(f->labels); f->labels = 0;}
	}
end:
	return r;
}
