/*
 interpreter.h

  Created: 2010.03.21
  Revised: 2011.09.29

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

2011.09.17
 nixed array16/32 codes and moved list from xF3 to xF1.  project calls for lean code.
 x4-x7 are now binary.
*/

#ifndef __INTERPRETER_H
#include "../enviro64.h"

/* ---- enums */

enum
{ /* size codes: size is 1<<(z&3), or c+1<<4 if _zc_list */
	_zc_list	= 0x7FFFFFF1, /* list_item, pos count (16 bytes each) */
	_zc_float,
	_zc_double,
	_zc_bin8,	/* x4 - binary (no specific format) */
	_zc_bin16,
	_zc_bin32,
	_zc_bin64,
	_zc_int8,	/* x8 - bit #3 = integer */
	_zc_int16,
	_zc_int32,
	_zc_int64,
	_zc_uint8,	/* xC */
	_zc_uint16,
	_zc_uint32,
	_zc_uint64	/* xF */
};

/* ---- typedefs/structures */

#if EVAL_FLOATING
typedef double		evalvalue;
typedef double		evalparse;
 #define EVAL_HUGEVAL			(evalvalue)((1L<<22)*(evalvalue)(1L<<30)-1)
 #define parse_eval(s,e,i)		parse_number(s,e,i,1)
 #define zc_eval			_zc_double
#else
typedef int32		evalvalue;
typedef long int	evalparse;
 #define EVAL_HUGEVAL			((evalvalue)1<<(sizeof(evalvalue)<<3)-1)
 #define parse_eval(s,e,i)		readanyinteger(s,e-s,i)
 #define zc_eval			_zc_int32
#endif

typedef union item_data/*(8)*/
{
	int64		i64;
	evalvalue	val, *vptr;
	char		str[8], *sptr;
	uint8		raw[8], *rptr;
	void		*ptr;

	struct list_item	*lptr;
} item_data;

/* non-reference pointers are mallocs when >=8 bytes */
#define ITEM_HAS_MALLOC(l)		((uint32)(l)->size-8<_zc_float-8)
#define ITEM_IS_BIN(l)			((l)->pos&0x80000000U)
#define ITEM_IS_REF(l)			((l)->size&0x80000000U)
#define ITEM_IS_RAWSTR(l)		((uint32)(l)->size<9)
#define ITEM_IS_RAWVAL(l)		((l)->size>=_zc_float)
#define ITEM_IS_MALLOC_REF(l)		((l)->size<(_zc_float|-1<<31))

/*
bin ref	low-size-high 
  0 0	str	 val
  0 1  sptr	 vptr
  1 0  func   stored MSB
  1 1  func  (native code)

 str refs point to item; if pos>0 then list, otherwise string
 non-static str ref to static str (func when both are on) to make const str.
*/
typedef struct list_item/*(16)*/
{
	int32		pos; /* position; bit 31 = static binary (fixed ordering) */
	int32		size; /* see size codes above; bit 31 = reference */
	item_data	data;
} list_item;

/* item #0 defines the terms to the list; pos is count (>0) */
/* names strlist: hash(first char in low 6 bits), Pascal string */
typedef list_item	*list_t;

/* ---- prototypes */

/* system standard */

int strcasecmp(const char *a,
	const char *b);
int strncasecmp(const char *a,
	const char *b,
	size_t z);

/* extended standard */

/* parses ASCII-compatible chars to return length of an entire line; 2011.08.13 */
/* escape char can be anything, such as '\', or 0 (null) for no escape char */
long sreadline(const char *p,
	long length,
	char escape);
/* same as sallocline, but uses fgetc/ungetc, allocates string w/ reallocfunc */
/* written output is capped to 0x7FFFFFF0 bytes */
char *freadline(struct FILE*,
	uint32 *length, /* in/out; can be NULL (C-string result) */
	void *(reallocfunc)(void*, /* can be NULL (uses realloc) */
		size_t),
	uint32 *lecount, /* line-ending count, in/out; can be NULL */
	int escape); /* can be 0 (no escape char) */

int makebasestring(void *outstring,
	const char *basechars, /* can be NULL (use default chars) */
	long value,
	unsigned base,
	int minlength,
	char sign); /* pass zero to interpret unsigned */

/* returns ptr to next character */
const char *readanyinteger(const char *str,
	long length,
	long int *outvalue);

const char *parse_number(const char *s,
	const char *e,
	double *v,
	uint8 r); /* real (take the fractional point) */

/* added; 2009.09.03 06:37 EDT */
/* negative result = error: ~position or memory (1<<31); overflow if completed */
long parse_textloop(const char *s,
	unsigned a,
	char **r, /* will realloc passed stored ptr; can be NULL */
	void (*b)(unsigned a, /* can be NULL */
		unsigned char v));

const char *parse_wordbreak(const char *s, /* does nothing if either are NULL */
	const char *e,
	uint8 *h); /* non-zero first char hash (can be NULL); 2011.09.15 */
const char *parse_rwordbreak(const char *s,
	const char *f);
const char *parse_skipwsp(const char *s,
	const char *e);
const char *parse_rskipwsp(const char *s,
	const char *f);

/* each division by zero will yield EVAL_HUGEVAL */
const char *parse_parameter(const char *s,
	long length,
	char *outstring, /* can be NULL; otherwise, expects at least 24 chars allocated */
	const char *(*getv)(const char *s,
		const char *e,
		evalvalue *i,
		void *u),
	const char *(*setv)(const char *s,
		const char *e,
		evalvalue i,
		void *u),
	void *userdata, /* for getv */
	evalvalue (*evalindir)(evalvalue)); /* indirection; (`@`); can be NULL */

/* the following exposed; 2011.09.16 */

const char *evaluate_line(const char *s,
	const char *e,
	evalvalue *v,
	const char *(*getv)( /* can be NULL */
		const char *s,
		const char *e,
		evalvalue*,
		void *u),
	const char *(*setv)( /* can be NULL */
		const char *s,
		const char *e,
		evalvalue,
		void *u),
	void *u,
	evalvalue (*evalindir)(evalvalue)); /* can be NULL */

/* lists */

/* if successful, returns the ptr to the new entry (beyond str item #0) */
list_item *list_add(list_t *p, /* can be NULL (new list) - item #1 still returned */
	long x,
	long z,
	void *v); /* data only copied in raw case */
void list_dispose(list_t p);

void *item_resize(list_item *l,
	long z);
long item_set(list_item *l,
	long x,
	long z,
	void *p);
double item_to_double(long x,
	long z,
	void *p);

uint32 getnamesitemindex(const uint8 *s,
	int32 sz,
	const char *n,
	uint8 nz,
	uint8 h); /* hash char from parse_wordbreak */

#define __INTERPRETER_H
#endif
